import pathlib as _pathlib
from importlib import metadata as _metadata

import toml as _toml

from ._clients import Client  # noqa
from ._clients import ClientError  # noqa
from ._mocking import Patch  # noqa
from ._sessions import Lobotomy  # noqa
from ._sessions import Session  # noqa
from ._cli import run as run_cli  # noqa

try:
    __version__ = _metadata.version(__package__)
except _metadata.PackageNotFoundError:  # pragma: no-cover
    # If the package is not installed such that it has distribution metadata
    # fallback to loading the version from the pyproject.toml file.
    __version__ = _toml.loads(
        _pathlib.Path(__file__)
        .parent.parent.joinpath('pyproject.toml')
        .read_text()
    )['tool']['poetry']['version']



import json
import pathlib
import typing

import toml
import yaml

from lobotomy import _fio
from lobotomy import _services
from lobotomy._cli import _definitions


def _get_path(
    context: '_definitions.CliContext',
) -> typing.Optional[pathlib.Path]:
    if context.args.configuration_file_path in [None, '-']:
        return None

    return (
        pathlib.Path(context.args.configuration_file_path)
        .expanduser()
        .absolute()
    )


def _mutate_service(
    service_data: dict, method: '_services.Method',
) -> typing.NoReturn:
    response_type = method.output.get('type', 'structure')
    new_response = method.configuration_output

    if method.name not in service_data:
        if response_type == 'list':
            # List entries confuse the side effect behavior of popping
            # responses off the list. To avoid that the list responses
            # are added as list items.
            service_data[method.name] = [new_response]
        else:
            service_data[method.name] = new_response
        return

    existing = service_data[method.name]
    if isinstance(existing, dict):
        service_data[method.name] = [existing, new_response]
    else:
        service_data[method.name] = existing + [new_response]


def run(context: '_definitions.CliContext') -> '_definitions.ExecutionResult':
    """Adds a new call to the specified command actions."""
    file_format = context.args.file_format
    path = _get_path(context)
    prefix = [item for item in (context.args.prefix or '').split('.') if item]

    try:
        configs = _fio.read(path, prefix, file_format=file_format)
    except (AttributeError, FileNotFoundError, TypeError):
        configs = {}

    service_name, method_name = context.args.boto_operation.split('.')
    service = _services.load_definition(service_name)
    method = service.lookup(method_name)

    if 'clients' not in configs:
        configs['clients'] = {}

    if service_name not in configs['clients']:
        configs['clients'][service_name] = {}

    _mutate_service(configs['clients'][service_name], method)

    if path:
        _fio.write(path, configs, prefix, file_format=file_format)
        return _definitions.ExecutionResult(
            code='ADDED', message='New call has been added to the configs.'
        ).echo()

    if file_format == 'json':
        print(json.dumps(configs, indent=2))
    elif file_format == 'toml':
        print(toml.dumps(configs))
    else:
        print(yaml.safe_dump(configs))

    return _definitions.ExecutionResult(
        code='ECHOED', message='New call has been echoed to stdout.'
    )



import datetime
import typing
from unittest.mock import MagicMock

import dateutil.parser
from botocore.response import StreamingBody


class InternalStreamer:
    """
    An internal class that replaces the HTTPResponse object that boto
    StreamingBody objects wrap. This is used to simulate the streaming
    body behaviors in returned streaming response objects.
    """

    def __init__(self, source: typing.AnyStr):
        """Wrap the source string in a streaming interface."""
        self._cursor = 0
        self._source = source

        # This exists for botocore compatibility setting timeouts.
        self._fp = MagicMock()

    def read(self, amt: int = None):
        """Simulates the streaming interface."""
        length = len(self._source)
        amount = amt if amt is not None else length
        start = max(0, self._cursor)
        end = min(length, self._cursor + amount)
        self._cursor = end
        return self._source[start:end]

    def close(self):
        # Exists only for compatibility cases.
        pass

    def __len__(self):
        return len(self._source)


def _cast_structure(definition: dict, value: dict) -> dict:
    """
    Converts a dictionary/structure botocore type into its formatted values
    by recursively casting all the members.

    :param definition:
        Specification definition for the associated value to cast.
    :param value:
        A loaded value to be cast into its boto client response value.
    :return:
        The cast version of the specified value that matches the format of
        the value as it would be returned in a boto client response.
    """
    return {k: cast(definition['members'].get(k), v) for k, v in value.items()}


def _cast_noop(definition: dict, value: typing.Any) -> typing.Any:
    """
    A non-casting, no-op operation used for unknown or uncastable value
    types.

    :param definition:
        Specification definition for the associated value to cast.
    :param value:
        A loaded value to be cast into its boto client response value.
    :return:
        An unmodified value as this is a no-op.
    """
    return value


def _cast_integer(definition: dict, value: typing.Any) -> int:
    """
    Converts a value into an integer value.

    :param definition:
        Specification definition for the associated value to cast.
    :param value:
        A loaded value to be cast into its boto client response value.
    :return:
        The cast version of the specified value that matches the format of
        the value as it would be returned in a boto client response.
    """
    return int(value)


def _cast_list(definition: dict, value: list) -> list:
    """
    Converts a list botocore type into its formatted values
    by recursively casting all the items.

    :param definition:
        Specification definition for the associated value to cast.
    :param value:
        A loaded value to be cast into its boto client response value.
    :return:
        The cast version of the specified value that matches the format of
        the value as it would be returned in a boto client response.
    """
    return [cast(definition.get('member'), v) for v in value]


def _cast_string(
    definition: dict, value: typing.Any,
) -> typing.Union[str, StreamingBody]:
    """
    Converts a string botocore type into its formatted value. If the
    definition indicates that the string is returned in streaming format,
    a StreamingBody object is returned instead with the string value
    injected into the body via an InternalStreamer object that mimics the
    actual streaming behavior of a real StreamingBody response.

    :param definition:
        Specification definition for the associated value to cast.
    :param value:
        A loaded value to be cast into its boto client response value.
    :return:
        The cast version of the specified value that matches the format of
        the value as it would be returned in a boto client response.
    """
    output = str(value)
    if definition.get('streaming'):
        output = StreamingBody(InternalStreamer(output), len(output))
    return output


def _cast_blob(
    definition: dict, value: typing.Any,
) -> typing.Union[bytes, StreamingBody]:
    """
    Converts a blob botocore type into its formatted value. If the
    definition indicates that the blob is returned in streaming format,
    a StreamingBody object is returned instead with the bytes value
    injected into the body via an InternalStreamer object that mimics the
    actual streaming behavior of a real StreamingBody response.

    :param definition:
        Specification definition for the associated value to cast.
    :param value:
        A loaded value to be cast into its boto client response value.
    :return:
        The cast version of the specified value that matches the format of
        the value as it would be returned in a boto client response.
    """
    output = str(value).encode()
    if definition.get('streaming'):
        output = StreamingBody(InternalStreamer(output), len(output))
    return output


def _cast_timestamp(
    definition: dict, value: typing.Union[float, int, str],
) -> datetime.datetime:
    """
    Converts a string date definition into a datetime as would be returned
    by the boto response. The dateutil library is used to parse the string
    for flexible representation. Additionally, an integer/float value can
    be specified, which will be treated as a unix timestamp. Note that
    boto returns timezone-aware datetime objects, so this loading process
    ensures that the returned datetime is timezone-aware. If no timezone
    was determined by the loading process, UTC is used by default.

    :param definition:
        Specification definition for the associated value to cast.
    :param value:
        A loaded value to be cast into its boto client response value.
    :return:
        The cast version of the specified value that matches the format of
        the value as it would be returned in a boto client response.
    """
    if isinstance(value, (int, float)):
        return datetime.datetime.fromtimestamp(value, tz=datetime.timezone.utc)

    output = dateutil.parser.parse(value)
    if not output.tzinfo:
        return output.astimezone(datetime.timezone.utc)
    return output


def cast(definition: typing.Optional[dict], value: typing.Any) -> typing.Any:
    """
    Casts the raw mocked value into its equivalent boto response type as
    defined by the definition object for the value. For complex types, this
    function acts recursively.

    :param definition:
        Specification definition for the associated value to cast.
    :param value:
        A loaded value to be cast into its boto client response value.
    :return:
        The cast version of the specified value that matches the format of
        the value as it would be returned in a boto client response.
    """
    if definition is None:
        # If there is no definition because it does not exist, just return
        # the value as-is.
        return value

    data_type = definition.get('type')
    conversions = {
        'structure': _cast_structure,
        'list': _cast_list,
        'timestamp': _cast_timestamp,
        'string': _cast_string,
        'long': _cast_integer,
        'integer': _cast_integer,
        'blob': _cast_blob,
    }
    caster = conversions.get(data_type, _cast_noop)
    output = caster(definition, value)
    return output
